
function twente(o)
% twente (-muenster symposium companion code)
% by Christian Himpe, 2015 ( http://wwwmath.uni-muenster.de/u/himpe )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*

if(exist('emgr')~=2)
    disp('emgr framework is required. Download at http://gramian.de/emgr.m');
    return;
end

%% Setup

J = 1;
N = 256;
O = J;
T = [0.0,0.01,1.0];
L = (T(3)-T(1))/T(2);
U = [ones(J,1)./T(2),zeros(J,L-1)];
X =  zeros(N,1);

rand('seed',1009);
A = -gallery('lehmer',N);
B = rand(N,J);
C = rand(O,N);

P = ones(N,1)*[0.1,0.25,0.50,0.75,0.9];
p = rand(N,1);

HYP = @(x,u,p) A*tanh(p.*x) + B*u;
OUT = @(x,u,p) C*x;

LIN = @(x,u,p) A*(p.*x) + B*u;

norm1 = @(y) T(2)*sum(abs(y(:)));
norm2 = @(y) sqrt(T(2)*dot(y(:),y(:)));
norm8 = @(y) max(y(:));

%% Main

Y = rk3(HYP,OUT,T,X,U,p); % Full Order

tic;
WC = emgr(HYP,OUT,[J,N,O],T,'c',P);
WO = emgr(HYP,OUT,[J,N,O],T,'o',P);
[UU,D,VV] = balance(WC,WO);
OFFLINE = toc

for I=1:N-1
    uu = VV(:,1:I);
    vv = UU(1:I,:);
    x = vv*X;
    hyp = @(x,u,p) vv*HYP(uu*x,u,p);
    out = @(x,u,p) OUT(uu*x,u,p);
    y = rk3(hyp,out,T,x,U,p); % Reduced Order
    l1b(I) = norm1(Y-y)/norm1(Y);
    l2b(I) = norm2(Y-y)/norm2(Y);
    l8b(I) = norm8(Y-y)/norm8(Y);
end;


tic;
WX = emgr(HYP,OUT,[J,N,O],T,'x',P);
[UU,D,VV] = svd(WX);
OFFLINE = toc

for I=1:N-1
    uu = UU(:,1:I);
    vv = uu';
    x = vv*X;
    hyp = @(x,u,p) vv*HYP(uu*x,u,p);
    out = @(x,u,p) OUT(uu*x,u,p);
    y = rk3(hyp,out,T,x,U,p); % Reduced Order
    l1n(I) = norm1(Y-y)/norm1(Y);
    l2n(I) = norm2(Y-y)/norm2(Y);
    l8n(I) = norm8(Y-y)/norm8(Y);
end;

tic;
WX = emgr(LIN,OUT,[J,N,O],T,'x',P);
[UU,D,VV] = svd(WX);
OFFLINE = toc

for I=1:N-1
    uu = UU(:,1:I);
    vv = uu';
    x = vv*X;
    lin = @(x,u,p) vv*LIN(uu*x,u,p);
    out = @(x,u,p) OUT(uu*x,u,p);
    y = rk3(lin,out,T,x,U,p); % Reduced Order
    l1l(I) = norm1(Y-y)/norm1(Y);
    l2l(I) = norm2(Y-y)/norm2(Y);
    l8l(I) = norm8(Y-y)/norm8(Y);
end;

%% Output

if(nargin==0), return; end

figure();
semilogy(1:N-1,l2b,'r','linewidth',2); hold on;
semilogy(1:N-1,l2n,'g','linewidth',2);
semilogy(1:N-1,l2l,'b','linewidth',2); hold off;
xlim([1,N-1]);
%ylim([10^floor(log10(min([l2b(:);l2n(:);l2l(:)]))-1),1]);
ylim([10^-5,1]);
pbaspect([2,1,1]);
legend('L_2 Error W_C W_O ','L_2 Error W_X ','L_2 Error W_X (linearized) ','location','northeast');
if(o==1), print('-depsc',[mfilename(),'.eps']); end;


%% ======== Balancer ========

function [X Y Z] = balance(WC,WO)

    [L D l] = svd(WC); LC = L*diag(sqrt(diag(D)));
    [L D l] = svd(WO); LO = L*diag(sqrt(diag(D)));
    [U Y V] = svd(LO'*LC);
    X = ( LO*U*diag(1.0./sqrt(diag(Y))) )';
    Z =   LC*V*diag(1.0./sqrt(diag(Y)));


%% ======== Integrator ========

function y = rk3(f,g,t,x,u,p)

    h = t(2);
    T = round(t(3)/h);

    y(:,1) = g(x,u(:,1),p);
    y(end,T) = 0;

    for t=1:T
        k1 = h*f(x,u(:,t),p);
        k2 = h*f(x + 0.5*k1,u(:,t),p);
        k3 = h*f(x + 0.75*k2,u(:,t),p);
        x = x + (2.0/9.0)*k1 + (1.0/3.0)*k2 + (4.0/9.0)*k3;
        y(:,t) = g(x,u(:,t),p);
    end;
